#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_system.h"
#include "esp_log.h"
#include "esp_wifi.h"
#include "esp_event.h"
#include "esp_event_loop.h"
#include "lwip/inet.h"

static const char tag[] = "MAIN";

#define STA_SSID			"DD-WRT"
#define STA_PASS			"12345678"

#define SOFT_AP_SSID		"ESP32_AP"
#define SOFT_AP_PASS		"testtest"

static tcpip_adapter_ip_info_t sta_ip_info = { 0, };
static tcpip_adapter_ip_info_t ap_ip_info = { 0, };

static esp_err_t event_handler(void *ctx, system_event_t *event) {
	switch (event->event_id) {
	case SYSTEM_EVENT_WIFI_READY:
		ESP_LOGI(tag, "Wifi ready")
		break;
	case SYSTEM_EVENT_STA_START:
		ESP_LOGI(tag, "STA: start")
		esp_wifi_connect();
		break;
	case SYSTEM_EVENT_STA_STOP:
		ESP_LOGI(tag, "STA: stop")
		break;
	case SYSTEM_EVENT_STA_CONNECTED:
		ESP_LOGI(tag, "STA: connected to SSID: %s (ch %u)", event->event_info.connected.ssid, event->event_info.connected.channel)
		break;
	case SYSTEM_EVENT_STA_GOT_IP:
		ESP_LOGI(tag, "STA: got IP")
		;
		break;
	case SYSTEM_EVENT_STA_DISCONNECTED:
		ESP_LOGI(tag, "STA: disconnected")
		break;
	case SYSTEM_EVENT_AP_START:
		ESP_LOGI(tag, "Soft-AP: start")
		break;
	case SYSTEM_EVENT_AP_STOP:
		ESP_LOGI(tag, "Soft-AP: stop")
		tcpip_adapter_dhcps_stop(TCPIP_ADAPTER_IF_AP);
		break;
	case SYSTEM_EVENT_AP_STACONNECTED:
		ESP_LOGI(tag, "Soft-AP: new client connected")
		tcpip_adapter_dhcp_status_t status;
		tcpip_adapter_dhcps_get_status(TCPIP_ADAPTER_IF_AP, &status);
		ESP_LOGI(tag, "DHCPS status %u", status)
		;
		break;
	case SYSTEM_EVENT_AP_STADISCONNECTED:
		ESP_LOGI(tag, "Soft-AP: client disconnected")
		break;
	default:
		ESP_LOGI(tag, "WiFi event - %u", event->event_id)
		break;
	}
	return ESP_OK;
}

bool fake_cfg_sta_ena(void) {
	return true;
}

bool fake_cfg_ap_ena(void) {
	return true;
}

void WifiReinit(void) {
	esp_wifi_stop();

	uint8_t mode = 0;
	if (fake_cfg_sta_ena()) {
		mode |= (1 << 0);
	}
	if (fake_cfg_ap_ena()) {
		mode |= (1 << 1);
	}

	wifi_mode_t newmode;

	switch (mode) {
	case 1:
		newmode = WIFI_MODE_STA;
		ESP_LOGI(tag, "WiFi set mode WIFI_MODE_STA")
		break;
	case 2:
		newmode = WIFI_MODE_AP;
		ESP_LOGI(tag, "WiFi set mode WIFI_MODE_AP")
		break;
	case 3:
		newmode = WIFI_MODE_APSTA;
		ESP_LOGI(tag, "WiFi set mode WIFI_MODE_APSTA")
		break;
	default:
		newmode = WIFI_MODE_NULL;
		ESP_LOGI(tag, "WiFi set mode WIFI_MODE_NULL")
		break;
	}

	esp_wifi_set_mode(newmode);

	if (mode & 1) {
		wifi_config_t sta_config = { .sta = { .bssid_set = false } };
		strncpy(sta_config.sta.ssid, STA_SSID, sizeof(sta_config.sta.ssid));
		strncpy(sta_config.sta.password, STA_PASS, sizeof(sta_config.sta.password));
		ESP_LOGI(tag, "sta.ssid: %s, sta.pass: %s", sta_config.sta.ssid, sta_config.sta.password);
		esp_wifi_set_config(WIFI_IF_STA, &sta_config);
	}

	if (mode & 2) {
		inet_aton("192.168.123.1", &ap_ip_info.ip);
		inet_aton("192.168.123.1", &ap_ip_info.gw);
		inet_aton("255.255.255.0", &ap_ip_info.netmask);

		tcpip_adapter_dhcps_stop(TCPIP_ADAPTER_IF_AP);
		ESP_ERROR_CHECK(tcpip_adapter_set_ip_info(TCPIP_ADAPTER_IF_AP, &ap_ip_info));
		ESP_ERROR_CHECK(tcpip_adapter_dhcps_start(TCPIP_ADAPTER_IF_AP));

		wifi_config_t ap_config = { .ap = { //
				//
						.ssid_len = 0,				//
						.channel = 0,				//
						.authmode = WIFI_AUTH_WPA2_PSK,				//
						.ssid_hidden = 0,				//
						.max_connection = 10, .beacon_interval = 100 } };

		strncpy(ap_config.ap.ssid, SOFT_AP_SSID, sizeof(ap_config.ap.ssid));
		strncpy(ap_config.ap.password, SOFT_AP_PASS, sizeof(ap_config.ap.password));
		ESP_LOGI(tag, "softap.ssid: %s, softap.pass: %s", ap_config.ap.ssid, ap_config.ap.password);
		esp_wifi_set_config(WIFI_IF_AP, &ap_config);
	}
	esp_wifi_start();
}

void app_main() {
	printf("\33[3J");
	ESP_LOGI(tag, "Start!");
	tcpip_adapter_init();
	ESP_ERROR_CHECK(esp_event_loop_init(event_handler, NULL));
	wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT()
	ESP_ERROR_CHECK(esp_wifi_init(&cfg));
	ESP_ERROR_CHECK(esp_wifi_set_storage(WIFI_STORAGE_RAM))
	WifiReinit();
	while (1)
		;
}
